from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class ChartTypesLanguageOptions(HighchartsMeta):
    """Chart type description strings.

    .. note::

      This is added to the chart information region.

      If there is only a single series type used in the chart, we use the format
      string for the series type, or default if missing. There is one format string
      for cases where there is only a single series in the chart, and one for multiple
      series of the same type.

    """

    def __init__(self, **kwargs):
        self._bar_multiple = None
        self._bar_single = None
        self._boxplot_multiple = None
        self._boxplot_single = None
        self._bubble_multiple = None
        self._bubble_single = None
        self._column_multiple = None
        self._column_single = None
        self._combination_chart = None
        self._default_multiple = None
        self._default_single = None
        self._empty_chart = None
        self._line_multiple = None
        self._line_single = None
        self._map_type_description = None
        self._pie_multiple = None
        self._pie_single = None
        self._scatter_multiple = None
        self._scatter_single = None
        self._spline_multiple = None
        self._spline_single = None
        self._unknown_map = None

        self.bar_multiple = kwargs.get('bar_multiple', None)
        self.bar_single = kwargs.get('bar_single', None)
        self.boxplot_multiple = kwargs.get('boxplot_multiple', None)
        self.boxplot_single = kwargs.get('boxplot_single', None)
        self.bubble_multiple = kwargs.get('bubble_multiple', None)
        self.bubble_single = kwargs.get('bubble_single', None)
        self.column_multiple = kwargs.get('column_multiple', None)
        self.column_single = kwargs.get('column_single', None)
        self.combination_chart = kwargs.get('combination_chart', None)
        self.default_multiple = kwargs.get('default_multiple', None)
        self.default_single = kwargs.get('default_single', None)
        self.empty_chart = kwargs.get('empty_chart', None)
        self.line_multiple = kwargs.get('line_multiple', None)
        self.line_single = kwargs.get('line_single', None)
        self.map_type_description = kwargs.get('map_type_description', None)
        self.pie_multiple = kwargs.get('pie_multiple', None)
        self.pie_single = kwargs.get('pie_single', None)
        self.scatter_multiple = kwargs.get('scatter_multiple', None)
        self.scatter_single = kwargs.get('scatter_single', None)
        self.spline_multiple = kwargs.get('spline_multiple', None)
        self.spline_single = kwargs.get('spline_single', None)
        self.unknown_map = kwargs.get('unknown_map', None)

    @property
    def bar_multiple(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_BAR_MULTIPLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bar_multiple

    @bar_multiple.setter
    def bar_multiple(self, value):
        if value == '':
            self._bar_multiple = ''
        else:
            self._bar_multiple = validators.string(value, allow_empty = True) or \
                constants.DEFAULT_LANG_ACS_CHART_TYPES_BAR_MULTIPLE

    @property
    def bar_single(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_BAR_SINGLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bar_single

    @bar_single.setter
    def bar_single(self, value):
        self._bar_single = validators.string(value, allow_empty = True)

    @property
    def boxplot_multiple(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_BOXPLOT_MULTIPLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._boxplot_multiple

    @boxplot_multiple.setter
    def boxplot_multiple(self, value):
        self._boxplot_multiple = validators.string(value, allow_empty = True)

    @property
    def boxplot_single(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_BOXPLOT_SINGLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._boxplot_single

    @boxplot_single.setter
    def boxplot_single(self, value):
        self._boxplot_single = validators.string(value, allow_empty = True)

    @property
    def bubble_multiple(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_BUBBLE_MULTIPLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bubble_multiple

    @bubble_multiple.setter
    def bubble_multiple(self, value):
        self._bubble_multiple = validators.string(value, allow_empty = True)

    @property
    def bubble_single(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_BUBBLE_SINGLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bubble_single

    @bubble_single.setter
    def bubble_single(self, value):
        self._bubble_single = validators.string(value, allow_empty = True)

    @property
    def column_multiple(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_COLUMN_MULTIPLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._column_multiple

    @column_multiple.setter
    def column_multiple(self, value):
        self._column_multiple = validators.string(value, allow_empty = True)

    @property
    def column_single(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_COLUMN_SINGLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._column_single

    @column_single.setter
    def column_single(self, value):
        self._column_single = validators.string(value, allow_empty = True)

    @property
    def combination_chart(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_COMBO}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._combination_chart

    @combination_chart.setter
    def combination_chart(self, value):
        self._combination_chart = validators.string(value, allow_empty = True)

    @property
    def default_multiple(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_DEFAULT_MULTIPLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._default_multiple

    @default_multiple.setter
    def default_multiple(self, value):
        self._default_multiple = validators.string(value, allow_empty = True)

    @property
    def default_single(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_DEFAULT_SINGLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._default_single

    @default_single.setter
    def default_single(self, value):
        self._default_single = validators.string(value, allow_empty = True)

    @property
    def empty_chart(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_EMPTY_CHART}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._empty_chart

    @empty_chart.setter
    def empty_chart(self, value):
        self._empty_chart = validators.string(value, allow_empty = True)

    @property
    def line_multiple(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_LINE_MULTIPLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._line_multiple

    @line_multiple.setter
    def line_multiple(self, value):
        self._line_multiple = validators.string(value, allow_empty = True)

    @property
    def line_single(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_LINE_SINGLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._line_single

    @line_single.setter
    def line_single(self, value):
        self._line_single = validators.string(value, allow_empty = True)

    @property
    def map_type_description(self) -> Optional[str]:
        """Defaults to
        ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_MAP_TYPE_DESCRIPTION}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._map_type_description

    @map_type_description.setter
    def map_type_description(self, value):
        self._map_type_description = validators.string(value, allow_empty = True)

    @property
    def pie_multiple(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_PIE_MULTIPLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._pie_multiple

    @pie_multiple.setter
    def pie_multiple(self, value):
        self._pie_multiple = validators.string(value, allow_empty = True)

    @property
    def pie_single(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_PIE_SINGLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._pie_single

    @pie_single.setter
    def pie_single(self, value):
        self._pie_single = validators.string(value, allow_empty = True)

    @property
    def scatter_multiple(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_SCATTER_MULTIPLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._scatter_multiple

    @scatter_multiple.setter
    def scatter_multiple(self, value):
        self._scatter_multiple = validators.string(value, allow_empty = True)

    @property
    def scatter_single(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_SCATTER_SINGLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._scatter_single

    @scatter_single.setter
    def scatter_single(self, value):
        self._scatter_single = validators.string(value, allow_empty = True)

    @property
    def spline_multiple(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_SPLINE_MULTIPLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._spline_multiple

    @spline_multiple.setter
    def spline_multiple(self, value):
        self._spline_multiple = validators.string(value, allow_empty = True)

    @property
    def spline_single(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_SPLINE_SINGLE}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._spline_single

    @spline_single.setter
    def spline_single(self, value):
        self._spline_single = validators.string(value, allow_empty = True)

    @property
    def unknown_map(self) -> Optional[str]:
        """Defaults to ``'{constants.DEFAULT_LANG_ACS_CHART_TYPES_UNKNOWN_MAP}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._unknown_map

    @unknown_map.setter
    def unknown_map(self, value):
        self._unknown_map = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'bar_multiple': as_dict.get('barMultiple', None),
            'bar_single': as_dict.get('barSingle', None),
            'boxplot_multiple': as_dict.get('boxplotMultiple', None),
            'boxplot_single': as_dict.get('boxplotSingle', None),
            'bubble_multiple': as_dict.get('bubbleMultiple', None),
            'bubble_single': as_dict.get('bubbleSingle', None),
            'column_multiple': as_dict.get('columnMultiple', None),
            'column_single': as_dict.get('columnSingle', None),
            'combination_chart': as_dict.get('combinationChart', None),
            'default_multiple': as_dict.get('defaultMultiple', None),
            'default_single': as_dict.get('defaultSingle', None),
            'empty_chart': as_dict.get('emptyChart', None),
            'line_multiple': as_dict.get('lineMultiple', None),
            'line_single': as_dict.get('lineSingle', None),
            'map_type_description': as_dict.get('mapTypeDescription', None),
            'pie_multiple': as_dict.get('pieMultiple', None),
            'pie_single': as_dict.get('pieSingle', None),
            'scatter_multiple': as_dict.get('scatterMultiple', None),
            'scatter_single': as_dict.get('scatterSingle', None),
            'spline_multiple': as_dict.get('splineMultiple', None),
            'spline_single': as_dict.get('splineSingle', None),
            'unknown_map': as_dict.get('unknownMap', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'barMultiple': self.bar_multiple,
            'barSingle': self.bar_single,
            'boxplotMultiple': self.boxplot_multiple,
            'boxplotSingle': self.boxplot_single,
            'bubbleMultiple': self.bubble_multiple,
            'bubbleSingle': self.bubble_single,
            'columnMultiple': self.column_multiple,
            'columnSingle': self.column_single,
            'combinationChart': self.combination_chart,
            'defaultMultiple': self.default_multiple,
            'defaultSingle': self.default_single,
            'emptyChart': self.empty_chart,
            'lineMultiple': self.line_multiple,
            'lineSingle': self.line_single,
            'mapTypeDescription': self.map_type_description,
            'pieMultiple': self.pie_multiple,
            'pieSingle': self.pie_single,
            'scatterMultiple': self.scatter_multiple,
            'scatterSingle': self.scatter_single,
            'splineMultiple': self.spline_multiple,
            'splineSingle': self.spline_single,
            'unknownMap': self.unknown_map
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.metaclasses import HighchartsMeta


class DataPointAccessibility(HighchartsMeta):
    """Accessibility options for a series."""

    def __init__(self, **kwargs):
        self._description = None
        self._enabled = None

        self.description = kwargs.get('description', None)
        self.enabled = kwargs.get('enabled', None)

    @property
    def description(self) -> Optional[str]:
        """Provide a description of the data point, announced to screen readers.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._description

    @description.setter
    def description(self, value):
        self._description = validators.string(value, allow_empty = True)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enable accessibility functionality for the data point.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'description': as_dict.get('description', None),
            'enabled': as_dict.get('enabled', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'description': self.description,
            'enabled': self.enabled,
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import validate_types
from highcharts_core.options.plot_options.bar import BarOptions
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class HistogramOptions(BarOptions):
    """General options to apply to all Histogram series types.

    A histogram is a column series which represents the distribution of the data set
    in the base series. Histogram splits data into bins and shows their frequencies.

    .. figure:: ../../../_static/histogram-example.png
      :alt: Histogram Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._bins_number = None
        self._bin_width = None

        self.bins_number = kwargs.get('bins_number', None)
        self.bin_width = kwargs.get('bin_width', None)

        super().__init__(**kwargs)

    @property
    def bins_number(self) -> Optional[CallbackFunction | str | int]:
        """A preferable number of bins. Defaults to ``'square-root'``.

        .. warning::

          This setting is a suggestion, so a histogram may have a different number of
          bins.

        By default, it is set to the square root of the base series' data length.

        Available options are:

          * ``'square-root'``
          * ``'sturges'``
          * ``'rice'``

        You can also provide a JavaScript function which takes a ``baseSeries`` as a
        parameter and returns a positive integer.

        :rtype: :class:`str <python:str>`, :class:`int <python:int>`, or
          :obj:`None <python:None>`
        """
        return self._bins_number

    @bins_number.setter
    def bins_number(self, value):
        if value is None:
            self._bins_number = None
        else:

            try:
                value = validate_types(value,
                                       types = CallbackFunction,
                                       allow_none = False)
            except (TypeError, ValueError):
                try:
                    value = validators.string(value)
                    value = value.lower()
                    if value not in ['square-root', 'sturges', 'rice']:
                        raise errors.HighchartsValueError(f'bins_number expects a number,'
                                                          f' a JavaScript callback '
                                                          f'function, or a string value '
                                                          f'of "square-root", "sturges", '
                                                          f'or "rice". Received: '
                                                          f'{value}')
                except TypeError:
                    value = validators.integer(value, minimum = 0)

            self._bins_number = value

    @property
    def bin_width(self) -> Optional[int | float | Decimal]:
        """Width of each bin. If :obj:`None <python:None>`, the default behavior is to
        calculate the width as ``(max - min) / number of bins``.

        .. note::

          This option takes precedence over :meth:`HistogramOptions.bins_number`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._bin_width

    @bin_width.setter
    def bin_width(self, value):
        self._bin_width = validators.numeric(value,
                                             allow_empty = True,
                                             minimum = 0)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'fill_color': as_dict.get('fillColor', None),
            'fill_opacity': as_dict.get('fillOpacity', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'negative_fill_color': as_dict.get('negativeFillColor', None),
            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'track_by_area': as_dict.get('trackByArea', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'center_in_category': as_dict.get('centerInCategory', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'colors': as_dict.get('colors', None),
            'grouping': as_dict.get('grouping', None),
            'group_padding': as_dict.get('groupPadding', None),
            'max_point_width': as_dict.get('maxPointWidth', None),
            'min_point_length': as_dict.get('minPointLength', None),
            'point_padding': as_dict.get('pointPadding', None),
            'point_range': as_dict.get('pointRange', None),
            'point_width': as_dict.get('pointWidth', None),

            'depth': as_dict.get('depth', None),
            'edge_color': as_dict.get('edgeColor', None),
            'edge_width': as_dict.get('edgeWidth', None),
            'group_z_padding': as_dict.get('groupZPadding', None),

            'bins_number': as_dict.get('binsNumber', None),
            'bin_width': as_dict.get('binWidth', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'binsNumber': self.bins_number,
            'binWidth': self.bin_width
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.options.plot_options.series import SeriesOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class HeatmapOptions(SeriesOptions):
    """General options to apply to all Heatmap series types.

    A heatmap is a graphical representation of data where the individual values
    contained in a matrix are represented as colors.

    .. warning::

      Heatmaps require that ``modules/heatmap`` is loaded client-side.

    .. figure:: ../../../_static/heatmap-example.png
      :alt: Heatmap Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._border_radius = None
        self._colsize = None
        self._interpolation = None
        self._null_color = None
        self._point_padding = None
        self._rowsize = None

        self.border_radius = kwargs.get('border_radius', None)
        self.colsize = kwargs.get('colsize', None)
        self.interpolation = kwargs.get('interpolation', None)
        self.null_color = kwargs.get('null_color', None)
        self.point_padding = kwargs.get('point_padding', None)
        self.rowsize = kwargs.get('rowsize', None)

        super().__init__(**kwargs)

    @property
    def border_radius(self) -> Optional[int | float | Decimal]:
        """The border radius for each heatmap item. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def colsize(self) -> Optional[int]:
        """The column size - how many X axis units each column in the heatmap should span.
        Defaults to ``1``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._colsize

    @colsize.setter
    def colsize(self, value):
        self._colsize = validators.integer(value,
                                           allow_empty = True,
                                           minimum = 1)

    @property
    def interpolation(self) -> Optional[bool]:
        """If ``True``, renders data points as an interpolated image.
        Defaults to ``False``.
        
        :rtype: :class:`bool <python:bool>`
        """
        return self._interpolation
    
    @interpolation.setter
    def interpolation(self, value):
        if value is None:
            self._interpolation = None
        else:
            self._interpolation = bool(value)

    @property
    def null_color(self) -> Optional[str | Gradient | Pattern]:
        """The color applied to null points. Defaults to ``'#f7f7f7'``.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._null_color

    @null_color.setter
    def null_color(self, value):
        from highcharts_core import utility_functions
        self._null_color = utility_functions.validate_color(value)

    @property
    def point_padding(self) -> Optional[int | float | Decimal]:
        """Padding between the points in the heatmap. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._point_padding

    @point_padding.setter
    def point_padding(self, value):
        self._point_padding = validators.numeric(value, allow_empty = True)

    @property
    def rowsize(self) -> Optional[int]:
        """The row size - how many Y axis units each heatmap row should span. Defaults to
        ``1``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._rowsize

    @rowsize.setter
    def rowsize(self, value):
        self._rowsize = validators.integer(value,
                                           allow_empty = True,
                                           minimum = 1)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'border_radius': as_dict.get('borderRadius', None),
            'colsize': as_dict.get('colsize', None),
            'interpolation': as_dict.get('interpolation', None),
            'null_color': as_dict.get('nullColor', None),
            'point_padding': as_dict.get('pointPadding', None),
            'rowsize': as_dict.get('rowsize', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderRadius': self.border_radius,
            'colsize': self.colsize,
            'interpolation': self.interpolation,
            'nullColor': self.null_color,
            'pointPadding': self.point_padding,
            'rowsize': self.rowsize
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class TilemapOptions(HeatmapOptions):
    """General options to apply to all Tilemap series types.

    A tilemap series is a type of heatmap where the tile shapes are configurable.

    .. tabs::

      .. tab:: Honeycomb Tilemap

        .. figure:: ../../../_static/tilemap-example.png
          :alt: Honeycomb Tilemap Example Chart
          :align: center

      .. tab:: Circle Tilemap

        .. figure:: ../../../_static/tilemap-example-circle.png
          :alt: Tilemap Example Chart
          :align: center

      .. tab:: Diamond Tilemap

        .. figure:: ../../../_static/tilemap-example-diamond.png
          :alt: Tilemap Example Chart
          :align: center

    """

    def __init__(self, **kwargs):
        self._tile_shape = None

        self.tile_shape = kwargs.get('tile_shape', None)

        super().__init__(**kwargs)

    @property
    def tile_shape(self) -> Optional[str]:
        """The shape of the tiles in the tilemap. Defaults to ``'hexagon'``.

        Possible values are:

          * ``'hexagon'``
          * ``'circle'``
          * ``'diamond'``
          * ``'square'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._tile_shape

    @tile_shape.setter
    def tile_shape(self, value):
        if not value:
            self._tile_shape = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['hexagon', 'circle', 'diamond', 'square']:
                raise errors.HighchartsValueError(f'tile_shape expects "hexagon", '
                                                  f'"circle", "diamond", or "square". '
                                                  f'Received: {value}')
            self._tile_shape = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'border_radius': as_dict.get('borderRadius', None),
            'colsize': as_dict.get('colsize', None),
            'interpolation': as_dict.get('interpolation', None),
            'null_color': as_dict.get('nullColor', None),
            'point_padding': as_dict.get('pointPadding', None),
            'rowsize': as_dict.get('rowsize', None),

            'tile_shape': as_dict.get('tileShape', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'tileShape': self.tile_shape
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.animation import AnimationOptions
from highcharts_core.utility_classes.data_labels import DataLabel
from highcharts_core.utility_classes.events import ClusterEvents
from highcharts_core.utility_classes.markers import Marker
from highcharts_core.utility_classes.states import States
from highcharts_core.utility_classes.zones import ClusterZone


class VectorLayoutAlgorithm(HighchartsMeta):
    """Options for the layout algorithm to apply to the Vector chart."""

    def __init__(self, **kwargs):
        self._distance = None
        self._grid_size = None
        self._iterations = None
        self._kmeans_threshold = None
        self._type = None

        self.distance = kwargs.get('distance', None)
        self.grid_size = kwargs.get('grid_size', None)
        self.iterations = kwargs.get('iterations', None)
        self.kmeans_threshold = kwargs.get('kmeans_threshold', None)
        self.type = kwargs.get('type', None)

    @property
    def distance(self) -> Optional[str | int | float | Decimal]:
        """When :meth:`VectorLayoutAlgorithm.type` is set to ``'kmeans'``, ``distance``
        is a maximum distance between point and cluster center so that this point will be
        inside the cluster. Defaults to ``40``.

        The distance is either a number expressed in pixels or a percentage defining a
        percentage of the plot area width.

        :rtype: :class:`str <python:str>`, numeric, or :obj:`None <python:None>`
        """
        return self._distance

    @distance.setter
    def distance(self, value):
        if value is None:
            self._distance = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise ValueError
            except ValueError:
                value = validators.numeric(value, minimum = 0)

            self._distance = value

    @property
    def grid_size(self) -> Optional[str | int | float | Decimal]:
        """When :meth:`VectorLayoutAlgorithm.type` is set to ``'grid'``, ``grid_size``
        is  is a size of a grid square element. Defaults to ``50``.

        The ``grid_size`` is either a number expressed in pixels or a percentage defining
        a percentage of the plot area width.

        :rtype: :class:`str <python:str>`, numeric, or :obj:`None <python:None>`
        """
        return self._grid_size

    @grid_size.setter
    def grid_size(self, value):
        if value is None:
            self._grid_size = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise ValueError
            except (TypeError, ValueError):
                value = validators.numeric(value, minimum = 0)

            self._grid_size = value

    @property
    def iterations(self) -> Optional[int]:
        """When :meth:`VectorLayoutAlgorithm.type` is set to ``'kmeans'``, the
        ``iterations`` indicates the number of times that the algorithm will be repeated
        to find cluster positions. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._iterations

    @iterations.setter
    def iterations(self, value):
        self._iterations = validators.integer(value,
                                              allow_empty = True,
                                              minimum = 1)

    @property
    def kmeans_threshold(self) -> Optional[int]:
        """When :meth:`VectorLayoutAlgorithm.type` is set to :obj:`None <python:None>` and
        the number of visible points is less than or equal to ``kmeans_threshold``, the
        ``'kmeans'`` algorithm is used by default to find clusters. When the number of
        visible points exceeds the ``kmeans_threshold``, the ``'grid'`` algorithm is used.
        Defaults to ``100``.

        .. hint::

          This threshold is a powerful tool to ensure good performance on large datasets
          and better cluster arrangemnet after zoom.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._kmeans_threshold

    @kmeans_threshold.setter
    def kmeans_threshold(self, value):
        self._kmeans_threshold = validators.integer(value,
                                                    allow_empty = True,
                                                    minimum = 0)

    @property
    def type(self) -> Optional[str]:
        """Type of algorithm to use to combine points into a cluster. Defaults to
        :obj:`None <python:None>`.

        There are three available algorithms:

          * ``'grid'`` - grid-based clustering technique. Points are assigned to squares
            of set size depending on their position on the plot area. Points inside the
            grid square are combined into a cluster. The grid size can be controlled by
            :meth:`VectorLayoutAlgorithm.grid_size` (grid size changes at certain zoom
            levels).
          * ``'kmeans'`` - based on K-Means clustering technique. In the first step,
            points are divided using the grid method (applying
            :meth:`VectorLayoutAlgorithm.distance` as the grid size) to find the initial
            amount of clusters. Next, each point is classified by computing the distance
            between each cluster center and that point. When the closest cluster distance
            is lower than :meth:`VectorLayoutAlgorithm.distance`, the point is added to
            this cluster otherwise is classified as noise. The algorithm is repeated until
            each cluster center does not change its previous position more than one pixel.
            This technique is more accurate but also more time consuming than the
            ``'grid'`` algorithm, especially for big datasets.
          * ``'optimizedKmeans'`` - based on K-Means clustering technique. This algorithm
            uses k-means algorithm only on the chart initialization or when chart extremes
            have greater range than on initialization. When a chart is redrawn the
            algorithm checks only clustered points distance from the cluster center and
            rebuilds it when the point is spaced enough to be outside the cluster. It
            provides a performance improvement and more stable clusters position so that
            it can be used rather on small and sparse datasets.

        When :obj:`None <python:None>`, the algorithm applied depends on
        :meth:`VectorLayoutAlgorithm.kmeans_threshold` such that if there are more visible
        points than the :meth:`kmeans_threshold <VectorLayoutAlgorithm.kmeans_threshold>`
        the ``'grid'`` algorithm is applied. If fewer, then ``'kmeans'`` is applied.

        A custom clustering algorithm can be applied by setting ``type`` to a
        JavaScript callback function. This function takes an array of ``processedXData``,
        ``processedYData``, ``processedXData`` indexes, and :class:`VectorLayoutAlgorithm`
        options as arguments and should return an object with grouped data.

        A custom algorithm should return an object similar to:

        .. code-block:: javascript

          {
           clusterId1: [{
               x: 573,
               y: 285,
               index: 1 // point index in the data array
           }, {
               x: 521,
               y: 197,
               index: 2
           }],
           clusterId2: [{
               ...
           }]
           ...
          }

        Where ``clusterId*`` (in the example above - a unique id of a cluster or noise) is
        an array of points belonging to a cluster. If the array has only one point or
        fewer points than set in :class:`Cluster.minimum_cluster_size`, it won't be
        combined into a cluster.

        """
        return self._type

    @type.setter
    def type(self, value):
        self._type = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'distance': as_dict.get('distance', None),
            'grid_size': as_dict.get('gridSize', None),
            'iterations': as_dict.get('iterations', None),
            'kmeans_threshold': as_dict.get('kmeansThreshold', None),
            'type': as_dict.get('type', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'distance': self.distance,
            'gridSize': self.grid_size,
            'iterations': self.iterations,
            'kmeansThreshold': self.kmeans_threshold,
            'type': self.type
        }

        return untrimmed


class ClusterOptions(HighchartsMeta):
    """Options for marker clusters, the concept of sampling the data values into
    larger blocks in order to ease readability and increase performance of the
    JavaScript charts.

    .. warning::

      The marker clusters module does not work with ``boost`` and ``draggable-points``
      modules.

    .. note::

      The marker clusters feature requires the ``marker-clusters.js`` file to be
      loaded, found in the modules directory of the download package, or online at
      `code.highcharts.com/modules/marker-clusters.js <code.highcharts.com/modules/marker-clusters.js>`_.

    """

    def __init__(self, **kwargs):
        self._allow_overlap = None
        self._animation = None
        self._data_labels = None
        self._drill_to_cluster = None
        self._enabled = None
        self._events = None
        self._layout_algorithm = None
        self._marker = None
        self._minimum_cluster_size = None
        self._states = None
        self._zones = None

        self.allow_overlap = kwargs.get('allow_overlap', None)
        self.animation = kwargs.get('animation', None)
        self.data_labels = kwargs.get('data_labels', None)
        self.drill_to_cluster = kwargs.get('drill_to_cluster', None)
        self.enabled = kwargs.get('enabled', None)
        self.events = kwargs.get('events', None)
        self.layout_algorithm = kwargs.get('layout_algorithm', None)
        self.marker = kwargs.get('marker', None)
        self.minimum_cluster_size = kwargs.get('minimum_cluster_size', None)
        self.states = kwargs.get('states', None)
        self.zones = kwargs.get('zones', None)

    @property
    def allow_overlap(self) -> Optional[bool]:
        """If ``True``, clusters are allowed to overlap. Otherwise, overlapping is
        prevented. Defaults to ``True``.

        .. warning::

          Preventing overlapping only works if
          :meth:`layout_algorithm.type <VectorLayoutAlgorithm.type>` is set to ``'grid'``.

        :returns: Flag indicating whether to allow clusters to overlap.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_overlap

    @allow_overlap.setter
    def allow_overlap(self, value):
        if value is None:
            self._allow_overlap = None
        else:
            self._allow_overlap = bool(value)

    @property
    def animation(self) -> Optional[AnimationOptions]:
        """Options for the cluster marker animation.

        :rtype: :class:`AnimationOptions` or :obj:`None <python:None>`
        """
        return self._animation

    @animation.setter
    @class_sensitive(AnimationOptions)
    def animation(self, value):
        self._animation = value

    @property
    def data_labels(self) -> Optional[DataLabel]:
        """Options for the cluster data labels.

        :rtype: :class:`DataLabel` or :obj:`None <python:None>`
        """
        return self._data_labels

    @data_labels.setter
    @class_sensitive(DataLabel)
    def data_labels(self, value):
        self._data_labels = value

    @property
    def drill_to_cluster(self) -> Optional[bool]:
        """If ``True``, zoom the plot area to the cluster points range when a cluster is
        clicked. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._drill_to_cluster

    @drill_to_cluster.setter
    def drill_to_cluster(self, value):
        if value is None:
            self._drill_to_cluster = None
        else:
            self._drill_to_cluster = bool(value)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enables the marker-clusters module. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def events(self) -> Optional[ClusterEvents]:
        """General event handlers for marker clusters.

        :rtype: :class:`SeriesEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(ClusterEvents)
    def events(self, value):
        self._events = value

    @property
    def layout_algorithm(self) -> Optional[VectorLayoutAlgorithm]:
        """Options for the layout algorithm to apply to the Vector chart.

        :rtype: :class:`VectorLayoutAlgorithm` or :obj:`None <python:None>`
        """
        return self._layout_algorithm

    @layout_algorithm.setter
    @class_sensitive(VectorLayoutAlgorithm)
    def layout_algorithm(self, value):
        self._layout_algorithm = value

    @property
    def marker(self) -> Optional[Marker]:
        """Options for the point markers of line-like series.

        Properties like ``fill_color``, ``line_color`` and ``line_width`` define the
        visual appearance of the markers. Other series types, like column series, don't
        have markers, but have visual options on the series level instead.

        :rtype: :class:`Marker` or :obj:`None <python:None>`
        """
        return self._marker

    @marker.setter
    @class_sensitive(Marker)
    def marker(self, value):
        self._marker = value

    @property
    def minimum_cluster_size(self) -> Optional[int]:
        """The minimum number of points to be combined into a cluster. Defaults to ``2``.

        .. note::

          This value has to be greater or equal to ``2``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._minimum_cluster_size

    @minimum_cluster_size.setter
    def minimum_cluster_size(self, value):
        self._minimum_cluster_size = validators.integer(value,
                                                        allow_empty = True,
                                                        minimum = 2)

    @property
    def states(self) -> Optional[States]:
        """Configuration for state-specific configuration to apply to all clusters.

        :rtype: :class:`States` or :obj:`None <python:None>`
        """
        return self._states

    @states.setter
    @class_sensitive(States)
    def states(self, value):
        self._states = value

    @property
    def zones(self) -> Optional[List[ClusterZone]]:
        """An array defining zones within marker clusters.

        :rtype: :obj:`None <python:None>` or :class:`list <python:list>` of
          :class:`ClusterZone` instances
        """
        return self._zones

    @zones.setter
    @class_sensitive(ClusterZone, force_iterable = True)
    def zones(self, value):
        self._zones = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'allow_overlap': as_dict.get('allowOverlap', None),
            'animation': as_dict.get('animation', None),
            'data_labels': as_dict.get('dataLabels', None),
            'drill_to_cluster': as_dict.get('drillToCluster', None),
            'enabled': as_dict.get('enabled', None),
            'events': as_dict.get('events', None),
            'layout_algorithm': as_dict.get('layoutAlgorithm', None),
            'marker': as_dict.get('marker', None),
            'minimum_cluster_size': as_dict.get('minimumClusterSize', None),
            'states': as_dict.get('states', None),
            'zones': as_dict.get('zones', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'allowOverlap': self.allow_overlap,
            'animation': self.animation,
            'dataLabels': self.data_labels,
            'drillToCluster': self.drill_to_cluster,
            'enabled': self.enabled,
            'events': self.events,
            'layoutAlgorithm': self.layout_algorithm,
            'marker': self.marker,
            'minimumClusterSize': self.minimum_cluster_size,
            'states': self.states,
            'zones': self.zones
        }

        return untrimmed



from typing import Optional
from highcharts_core.options.axes.y_axis import YAxis


class ParallelAxesOptions(YAxis):
    """Common options for all Y-axes rendered in a parallel coordinates plot.

    .. warning::

      This feature requires ``modules/parallel-coordinates.js``.

    """
    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'chart.parallelAxes'



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta, JavaScriptDict
from highcharts_core.options.plot_options.accessibility import TypeOptionsAccessibility
from highcharts_core.options.series.labels import SeriesLabel
from highcharts_core.options.plot_options.points import OnPointOptions
from highcharts_core.options.plot_options.points import Point
from highcharts_core.options.plot_options.sonification import SeriesSonification
from highcharts_core.options.tooltips import Tooltip
from highcharts_core.utility_classes.animation import AnimationOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.data_labels import DataLabel
from highcharts_core.utility_classes.events import SeriesEvents
from highcharts_core.utility_classes.markers import Marker
from highcharts_core.utility_classes.states import States
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class GenericTypeOptions(HighchartsMeta):
    """Generic class that is used as a base class for type-specific :class:`PlotOptions`.
    """

    def __init__(self, **kwargs):
        self._accessibility = None
        self._allow_point_select = None
        self._animation = None
        self._class_name = None
        self._clip = None
        self._color = None
        self._cursor = None
        self._custom = None
        self._dash_style = None
        self._data_labels = None
        self._description = None
        self._enable_mouse_tracking = None
        self._events = None
        self._include_in_data_export = None
        self._keys = None
        self._label = None
        self._legned_symbol = None
        self._linked_to = None
        self._marker = None
        self._on_point = None
        self._opacity = None
        self._point = None
        self._point_description_formatter = None
        self._selected = None
        self._show_checkbox = None
        self._show_in_legend = None
        self._skip_keyboard_navigation = None
        self._sonification = None
        self._states = None
        self._sticky_tracking = None
        self._threshold = None
        self._tooltip = None
        self._turbo_threshold = None
        self._visible = None

        self.accessibility = kwargs.get('accessibility', None)
        self.allow_point_select = kwargs.get('allow_point_select', None)
        self.animation = kwargs.get('animation', None)
        self.class_name = kwargs.get('class_name', None)
        self.clip = kwargs.get('clip', None)
        self.color = kwargs.get('color', None)
        self.cursor = kwargs.get('cursor', None)
        self.custom = kwargs.get('custom', None)
        self.dash_style = kwargs.get('dash_style', None)
        self.data_labels = kwargs.get('data_labels', None)
        self.description = kwargs.get('description', None)
        self.enable_mouse_tracking = kwargs.get('enable_mouse_tracking', None)
        self.events = kwargs.get('events', None)
        self.include_in_data_export = kwargs.get('include_in_data_export', None)
        self.keys = kwargs.get('keys', None)
        self.label = kwargs.get('label', None)
        self.legend_symbol = kwargs.get('legend_symbol', None)
        self.linked_to = kwargs.get('linked_to', None)
        self.marker = kwargs.get('marker', None)
        self.on_point = kwargs.get('on_point', None)
        self.opacity = kwargs.get('opacity', None)
        self.point = kwargs.get('point', None)
        self.point_description_formatter = kwargs.get('point_description_formatter', None)
        self.selected = kwargs.get('selected', None)
        self.show_checkbox = kwargs.get('show_checkbox', None)
        self.show_in_legend = kwargs.get('show_in_legend', None)
        self.skip_keyboard_navigation = kwargs.get('skip_keyboard_navigation', None)
        self.sonification = kwargs.get('sonification', None)
        self.states = kwargs.get('states', None)
        self.sticky_tracking = kwargs.get('sticky_tracking', None)
        self.threshold = kwargs.get('threshold', None)
        self.tooltip = kwargs.get('tooltip', None)
        self.turbo_threshold = kwargs.get('turbo_threshold', None)
        self.visible = kwargs.get('visible', None)

        super().__init__(**kwargs)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return f'plotOptions.{self.type}'

    @property
    def type(self) -> str:
        """Indicates the type of series that is represented by this instance.

        .. warning::

          This proprety is read-only!

        :rtype: :class:`str <python:str>`
        """
        class_name = self.__class__.__name__
        class_name = class_name.replace('TypeOptions', '')
        class_name = class_name.replace('Options', '')
        if class_name.endswith('Series') and class_name != 'Series':
            class_name = class_name.replace('Series', '')

        return class_name.lower()

    @type.setter
    def type(self, value):
        raise errors.HighchartsReadOnlyError('type is a read-only property and cannot be '
                                             'set manually')

    @property
    def accessibility(self) -> Optional[TypeOptionsAccessibility]:
        """Accessibility options for a series.

        :rtype: :class:`TypeOptionsAccessibility` or :obj:`None <python:None>`
        """
        return self._accessibility

    @accessibility.setter
    @class_sensitive(TypeOptionsAccessibility)
    def accessibility(self, value):
        self._accessibility = value

    @property
    def allow_point_select(self) -> Optional[bool]:
        """Allow this series' points to be selected by clicking on the graphic (columns,
        point markers, pie slices, map areas etc).

        The selected points can be handled in JavaScript by point select and unselect
        events, or collectively by the (JavaScript) ``getSelectedPoints()`` function.

        And alternative way of selecting points is through dragging.

        Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_point_select

    @allow_point_select.setter
    def allow_point_select(self, value):
        if value is None:
            self._allow_point_select = None
        else:
            self._allow_point_select = bool(value)

    @property
    def animation(self) -> Optional[AnimationOptions]:
        """Enable or disable the initial animation when a series is displayed.

        The animation can also be set as a configuration object. Please note that this
        option only applies to the initial animation of the series itself. For other
        animations, see :class:`Chart.animation` and the ``animation`` parameter under the
        (JavaScript) API methods. The following properties are supported:

          * ``defer``: The animation delay time in milliseconds.
          * ``duration``: The duration of the animation in milliseconds.
          * ``easing``: Can be a string reference to an easing function set on the Math
            object or a function.

        .. warning::

          Due to poor performance, animation is disabled in old IE browsers for several
          chart types.

        :rtype: :class:`AnimationOptions` or :obj:`None <python:None>`
        """
        return self._animation

    @animation.setter
    @class_sensitive(AnimationOptions)
    def animation(self, value):
        self._animation = value

    @property
    def class_name(self) -> Optional[str]:
        """The additional CSS class name to apply to the series' graphical elements.

        .. note::

          This option is additive to the default class names - it does not replace them.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def clip(self) -> Optional[bool]:
        """If ``False``, allows the series to be rendered in the entire plot area. If
        ``True``, constrains where the series can be rendered within the plot area.
        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._clip

    @clip.setter
    def clip(self, value):
        if value is None:
            self._clip = None
        else:
            self._clip = bool(value)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The main color of the series.

        In line type series it applies to the line and the point markers unless otherwise
        specified. In bar type series it applies to the bars unless a color is specified
        per point. The default value is pulled from the :meth:`Options.colors` array.

        :returns: The main color applied to the series.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def cursor(self) -> Optional[str]:
        """The style of cursor to use when the user's mouse hovers over the data series.

        Acceptable values are:

          * ``'alias'``
          * ``'all-scroll'``
          * ``'auto'``
          * ``'cell'``
          * ``'col-resize'``
          * ``'context-menu'``
          * ``'copy'``
          * ``'crosshair'``
          * ``'default'``
          * ``'e-resize'``
          * ``'ew-resize'``
          * ``'grab'``
          * ``'grabbing'``
          * ``'help'``
          * ``'move'``
          * ``'n-resize'``
          * ``'ne-resize'``
          * ``'nesw-resize'``
          * ``'no-drop'``
          * ``'none'``
          * ``'not-allowed'``
          * ``'ns-resize'``
          * ``'nw-resize'``
          * ``'nwse-resize'``
          * ``'pointer'``
          * ``'progress'``
          * ``'row-resize'``
          * ``'s-resize'``
          * ``'se-resize'``
          * ``'sw-resize'``
          * ``'text'``
          * ``'vertical-text'``
          * ``'w-resize'``
          * ``'wait'``
          * ``'zoom-in'``
          * ``'zoom-out'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._cursor

    @cursor.setter
    def cursor(self, value):
        if not value:
            self._cursor = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in constants.SUPPORTED_CURSOR_VALUES:
                raise errors.HighchartsValueError(f'cursor expects a valid cursor value. '
                                                  f'Received: {value}')
            self._cursor = value

    @property
    def custom(self) -> Optional[JavaScriptDict]:
        """A reserved subspace to store options and values for customized functionality.

        Here you can add additional data for your own event callbacks and formatter
        callbacks.

        :rtype: :class:`dict <python:dict>` or :obj:`None <python:None>`
        """
        return self._custom

    @custom.setter
    @class_sensitive(JavaScriptDict)
    def custom(self, value):
        self._custom = value

    @property
    def dash_style(self) -> Optional[str]:
        """Name of the dash style to use for the graph, or for some series types the
        outline of each shape.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._dash_style

    @dash_style.setter
    def dash_style(self, value):
        if not value:
            self._dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'dash_style expects a recognized value'
                                                  f', but received: {value}')
            self._dash_style = value

    @property
    def data_labels(self) -> Optional[DataLabel | List[DataLabel]]:
        """Options for the series data labels, appearing next to each data point.

        .. note::

          To have multiple data labels per data point, you can also supply a collection of
          :class:`DataLabel` configuration settings.

        :rtype: :class:`DataLabel`, :class:`list <python:list>` of :class:`DataLabel`, or
          :obj:`None <python:None>`
        """
        return self._data_labels

    @data_labels.setter
    def data_labels(self, value):
        if not value:
            self._data_labels = None
        else:
            if checkers.is_iterable(value):
                self._data_labels = validate_types(value,
                                                   types = DataLabel,
                                                   allow_none = False,
                                                   force_iterable = True)
            else:
                self._data_labels = validate_types(value,
                                                   types = DataLabel,
                                                   allow_none = False)

    @property
    def description(self) -> Optional[str]:
        """A description of the series to add to the screen reader information about the
        series.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._description

    @description.setter
    def description(self, value):
        self._description = validators.string(value, allow_empty = True)

    @property
    def enable_mouse_tracking(self) -> Optional[bool]:
        """If ``True``, enables mouse tracking for the series (used to capture point
        tooltips, click events on graphs and points, etc.). If ``False``, disables
        mouse tracking for the series (which can help performance). Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enable_mouse_tracking

    @enable_mouse_tracking.setter
    def enable_mouse_tracking(self, value):
        if value is None:
            self._enable_mouse_tracking = None
        else:
            self._enable_mouse_tracking = bool(value)

    @property
    def events(self) -> Optional[SeriesEvents]:
        """General event handlers for the series items.

        .. note::

          These event hooks can also be attached to the series at run time using the
          (JavaScript) ``Highcharts.addEvent()`` function.

        :rtype: :class:`SeriesEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(SeriesEvents)
    def events(self, value):
        self._events = value

    @property
    def include_in_data_export(self) -> Optional[bool]:
        """If ``False``, will prevent the data series from being included in any form of
        data export. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._include_in_data_export

    @include_in_data_export.setter
    def include_in_data_export(self, value):
        if value is None:
            self._include_in_data_export = None
        else:
            self._include_in_data_export = bool(value)

    @property
    def keys(self) -> Optional[List[str]]:
        """An array specifying which option maps to which key in the data point array.

        This makes it convenient to work with unstructured data arrays from different
        sources.

        :rtype: :class:`list <python:list>` of :class:`str <python:str>`, or
          :obj:`None <python:None>`
        """
        return self._keys

    @keys.setter
    def keys(self, value):
        if not value:
            self._keys = None
        else:
            self._keys = [validators.string(x)
                          for x in validators.iterable(value)]

    @property
    def label(self) -> Optional[SeriesLabel]:
        """Series labels are placed as close to the series as possible in a natural way,
        seeking to avoid other series. The goal of this feature is to make the chart more
        easily readable, like if a human designer placed the labels in the optimal
        position.

        .. note::

          The series labels currently work with series types having a graph or an area.

        :rtype: :class:`SeriesLabel` or :obj:`None <python:None>`
        """
        return self._label

    @label.setter
    @class_sensitive(SeriesLabel)
    def label(self, value):
        self._label = value

    @property
    def legend_symbol(self) -> Optional[str]:
        """The type of legend symbol to render for the series. Accepts either 
        ``'lineMarker'`` or ``'rectangle'``. Defaults to ``'rectangle'``.
        
        :rtype: :class:`str <python:str>`
        """
        return self._legend_symbol
    
    @legend_symbol.setter
    def legend_symbol(self, value):
        if not value:
            self._legend_symbol = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value == 'linemarker':
                value = 'lineMarker'
            if value not in ['lineMarker', 'rectangle']:
                raise errors.HighchartsValueError(f'legend_symbol expects either '
                                                  f'"lineMarker" or "rectangle". '
                                                  f'Received: "{value}".')
            self._legend_symbol = value

    @property
    def linked_to(self) -> Optional[str]:
        """The id of another series to link to.

        .. hint::

          The value can be ``':previous'`` to link to the previous series. When two series
          are linked, only the first one appears in the legend. Toggling the visibility of
          this also toggles the linked series.

        .. note::

          If the master series uses data sorting and linked series does not have its own
          sorting definition, the linked series will be sorted in the same order as the
          master one.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._linked_to

    @linked_to.setter
    def linked_to(self, value):
        self._linked_to = validators.string(value, allow_empty = True)

    @property
    def marker(self) -> Optional[Marker]:
        """Options for the point markers of line-like series.

        Properties like ``fill_color``, ``line_color`` and ``line_width`` define the
        visual appearance of the markers. Other series types, like column series, don't
        have markers, but have visual options on the series level instead.

        :rtype: :class:`Marker` or :obj:`None <python:None>`
        """
        return self._marker

    @marker.setter
    @class_sensitive(Marker)
    def marker(self, value):
        self._marker = value

    @property
    def on_point(self) -> Optional[OnPointOptions]:
        """Options for the Series on point feature, which is currently only supported by
        ``pie`` and ``sunburst`` chargs.

        :rtype: :class:`OnPointOptions` or :obj:`None <python:None>`
        """
        return self._on_point

    @on_point.setter
    @class_sensitive(OnPointOptions)
    def on_point(self, value):
        self._on_point = value

    @property
    def opacity(self) -> Optional[float]:
        """Opacity of a series parts: line, fill (e.g. area), and labels.

        :rtype: :class:`float <python:float>`
        """
        return self._opacity

    @opacity.setter
    def opacity(self, value):
        self._opacity = validators.float(value,
                                         allow_empty = True,
                                         minimum = 0.0,
                                         maximum = 1.0)

    @property
    def point(self) -> Optional[Point]:
        """Properties for each single point.

        :rtype: :class:`Point` or :obj:`None <python:None>`
        """
        return self._point

    @point.setter
    @class_sensitive(Point)
    def point(self, value):
        self._point = value

    @property
    def point_description_formatter(self) -> Optional[CallbackFunction]:
        """Same as for :meth:`Accessibility.series.description_formatter`, only for an
        individual series. Overrides the chart-wide configuration.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._point_description_formatter

    @point_description_formatter.setter
    @class_sensitive(CallbackFunction)
    def point_description_formatter(self, value):
        self._point_description_formatter = value

    @property
    def selected(self) -> Optional[bool]:
        """If ``True``, the series is selected initially (by default, without user
        interaction). Defaults to ``False``.

        .. note::

          If :meth:`GenericTypeOptions.show_checkbox` is ``True``, then the checkbox
          will be checked if ``selected`` is ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._selected

    @selected.setter
    def selected(self, value):
        if value is None:
            self._selected = None
        else:
            self._selected = bool(value)

    @property
    def show_checkbox(self) -> Optional[bool]:
        """If ``True``, a checkbox is displayed next to the legend item to allow selecting
        the series.

        .. note::

          The state of the checkbox is controlled by the
          :meth:`GenericTypeOptions.selected` property.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_checkbox

    @show_checkbox.setter
    def show_checkbox(self, value):
        if value is None:
            self._show_checkbox = None
        else:
            self._show_checkbox = bool(value)

    @property
    def show_in_legend(self) -> Optional[bool]:
        """Whether to display this particular series or series type in the legend.
        Standalone series are shown in the legend by default, and linked series are not.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_in_legend

    @show_in_legend.setter
    def show_in_legend(self, value):
        if value is None:
            self._show_in_legend = None
        else:
            self._show_in_legend = bool(value)

    @property
    def skip_keyboard_navigation(self) -> Optional[bool]:
        """If ``True``, the accessibility module will skip past this series when executing
        keyboard navigation.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._skip_keyboard_navigation

    @skip_keyboard_navigation.setter
    def skip_keyboard_navigation(self, value):
        if value is None:
            self._skip_keyboard_navigation = None
        else:
            self._skip_keyboard_navigation = bool(value)

    @property
    def sonification(self) -> Optional[SeriesSonification]:
        """Sonification configuration for the series type/series.
        
        :rtype: :class:`SeriesSonification <highcharts_core.options.plot_options.sonification.SeriesSonification>` or
          :obj:`None <python:None>`
        """
        return self._sonification
    
    @sonification.setter
    @class_sensitive(SeriesSonification)
    def sonification(self, value):
        self._sonification = value

    @property
    def states(self) -> Optional[States]:
        """Configuration for state-specific configuration to apply to the data series.

        :rtype: :class:`States` or :obj:`None <python:None>`
        """
        return self._states

    @states.setter
    @class_sensitive(States)
    def states(self, value):
        self._states = value

    @property
    def sticky_tracking(self) -> Optional[bool]:
        """Sticky tracking of mouse events.

        When ``True``, the (JavaScript) ``mouseOut`` event on a series is not triggered
        until the mouse moves over another series, or out of the plot area.

        When ``False``, the (JavaScript) ``mouseOut`` event on a series is triggered when
        the mouse leaves the area around the series' graph or markers. This also implies
        the tooltip when not shared.

        When ``False`` and :meth:`PlotOptions.tooltip.shared` is also ``False``, the
        tooltip will be hidden when moving the mouse between series.

        Defaults to ``True`` for line and area type series, but to ``False`` for columns,
        pies, etc.

        .. note::

          The boost module will force this option because of technical limitations.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._sticky_tracking

    @sticky_tracking.setter
    def sticky_tracking(self, value):
        if value is None:
            self._sticky_tracking = None
        else:
            self._sticky_tracking = bool(value)

    @property
    def threshold(self) -> Optional[constants.EnforcedNullType | int | float | Decimal]:
        """The Y axis value to serve as the base for the columns, for distinguishing
        between values above and below a threshold. Defaults to ``0``.

        If :class:`EnforcedNullType`, the columns extend from the padding Y axis minimum.

        :rtype: numeric or :class:`EnforcedNullType` or :obj:`None <python:None>`
        """
        return self._threshold

    @threshold.setter
    def threshold(self, value):
        if value == constants.EnforcedNull:
            self._threshold = constants.EnforcedNull
        else:
            self._threshold = validators.numeric(value, allow_empty = True)

    @property
    def tooltip(self) -> Optional[Tooltip]:
        """A configuration object for the tooltip rendering of each single series.
        Properties are inherited from tooltip, but only the following properties can be
        defined on a series level.

        :rtype: :class:`Tooltip` or :obj:`None <python:None>`
        """
        return self._tooltip

    @tooltip.setter
    @class_sensitive(Tooltip)
    def tooltip(self, value):
        self._tooltip = value

    @property
    def turbo_threshold(self) -> Optional[int]:
        """When a series contains a data array longer than this value, only one
        dimensional arrays of numbers, or two dimensional arrays with x and y values are
        allowed. Also, only the first point is tested, and the rest are assumed to be the
        same format. This saves expensive data checking and indexing in long series.
        Set it to ``0`` or :obj:`None <python:None>` to disable.

        Defaults to ``1000``.

        .. note::

          In boost mode, turbo threshold is forced. Only array of numbers or two
          dimensional arrays are allowed.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._turbo_threshold

    @turbo_threshold.setter
    def turbo_threshold(self, value):
        self._turbo_threshold = validators.integer(value,
                                                   allow_empty = True,
                                                   minimum = 0)

    @property
    def visible(self) -> Optional[bool]:
        """If ``True``, the series is initially visible. If ``False``, the series is
        hidden by default. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._visible

    @visible.setter
    def visible(self, value):
        if value is None:
            self._visible = None
        else:
            self._visible = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'accessibility': self.accessibility,
            'allowPointSelect': self.allow_point_select,
            'animation': self.animation,
            'className': self.class_name,
            'clip': self.clip,
            'color': self.color,
            'cursor': self.cursor,
            'custom': self.custom,
            'dashStyle': self.dash_style,
            'dataLabels': self.data_labels,
            'description': self.description,
            'enableMouseTracking': self.enable_mouse_tracking,
            'events': self.events,
            'includeInDataExport': self.include_in_data_export,
            'keys': self.keys,
            'label': self.label,
            'legendSymbol': self.legend_symbol,
            'linkedTo': self.linked_to,
            'marker': self.marker,
            'onPoint': self.on_point,
            'opacity': self.opacity,
            'point': self.point,
            'pointDescriptionFormatter': self.point_description_formatter,
            'selected': self.selected,
            'showCheckbox': self.show_checkbox,
            'showInLegend': self.show_in_legend,
            'skipKeyboardNavigation': self.skip_keyboard_navigation,
            'sonification': self.sonification,
            'states': self.states,
            'stickyTracking': self.sticky_tracking,
            'threshold': self.threshold,
            'tooltip': self.tooltip,
            'turboThreshold': self.turbo_threshold,
            'visible': self.visible,
            'type': self.type
        }

        return untrimmed



"""Tests for ``highcharts.boost``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.boost import Boost as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'allow_force': True,
      'debug': {
          'showSkipSummary': False,
          'timeBufferCopy': True,
          'timeKDTree': True,
          'timeRendering': False,
          'timeSeriesProcessing': True,
          'timeSetup': True
      },
      'enabled': True,
      'pixel_ratio': 2,
      'series_threshold': 5000,
      'use_gpu_translations': True,
      'use_preallocated': True
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('boost/01.js', False, None),

    ('boost/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('boost/01.js', True, None),

    ('boost/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.options.annotations.points import AnnotationPoint
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.shadows import ShadowOptions
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class AnnotationLabelOptionAccessibility(HighchartsMeta):
    """Accessibility options applied to an annotation label."""

    def __init__(self, **kwargs):
        self._description = None

        self.description = kwargs.get('description', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'annotations.labelOptions.accessibility'

    @property
    def description(self) -> Optional[str]:
        """Description of an annotation label for screen readers and other assistive
        technology.

        :returns: Description of an annotation label
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._description

    @description.setter
    def description(self, value):
        self._description = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'description': as_dict.get('description', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'description': self.description
        }

        return untrimmed


class LabelOptions(HighchartsMeta):
    """Options applied to all annotation labels."""

    def __init__(self, **kwargs):
        self._accessibility = None
        self._align = None
        self._allow_overlap = None
        self._background_color = None
        self._border_color = None
        self._border_radius = None
        self._border_width = None
        self._class_name = None
        self._crop = None
        self._distance = None
        self._format = None
        self._formatter = None
        self._include_in_data_export = None
        self._overflow = None
        self._padding = None
        self._shadow = None
        self._shape = None
        self._style = None
        self._use_html = None
        self._vertical_align = None
        self._x = None
        self._y = None

        self.accessibility = kwargs.get('accessibility', None)
        self.align = kwargs.get('align', None)
        self.allow_overlap = kwargs.get('allow_overlap', None)
        self.background_color = kwargs.get('background_color', None)
        self.border_color = kwargs.get('border_color', None)
        self.border_radius = kwargs.get('border_radius', None)
        self.border_width = kwargs.get('border_width', None)
        self.class_name = kwargs.get('class_name', None)
        self.crop = kwargs.get('crop', None)
        self.distance = kwargs.get('distance', None)
        self.format = kwargs.get('format', None)
        self.formatter = kwargs.get('formatter', None)
        self.include_in_data_export = kwargs.get('include_in_data_export', None)
        self.overflow = kwargs.get('overflow', None)
        self.padding = kwargs.get('padding', None)
        self.shadow = kwargs.get('shadow', None)
        self.shape = kwargs.get('shape', None)
        self.style = kwargs.get('style', None)
        self.text = kwargs.get('text', None)
        self.use_html = kwargs.get('use_html', None)
        self.vertical_align = kwargs.get('vertical_align', None)
        self.x = kwargs.get('x', None)
        self.y = kwargs.get('y', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'annotations.labelOptions'

    @property
    def accessibility(self) -> Optional[AnnotationLabelOptionAccessibility]:
        """Accessibility options applied to an annotation label.

        :returns: Accessibility options applied to an annotation label.
        :rtype: :class:`AnnotationLabelOptionAccessibility` or :obj:`None <python:None>`
        """
        return self._accessibility

    @accessibility.setter
    @class_sensitive(AnnotationLabelOptionAccessibility)
    def accessibility(self, value):
        self._accessibility = value

    @property
    def align(self) -> Optional[str]:
        """The alignment of the annotation's label. Defaults to
        ``'center'``.

        Accepts:

          * ``'left'``
          * ``'center'``
          * ``'right'``

        .. hint::

          If right, the right side of the label should be touching the point.

        :returns: The alignment of the annotation's label.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._align

    @align.setter
    def align(self, value):
        if not value:
            self._align = None
        else:
            value = validators.string(value, allow_empty = False)
            value = value.lower()
            if value not in ['left', 'center', 'right']:
                raise errors.HighchartsValueError(f'align must be either "left", '
                                                  f'"center", or "right". Was: {value}')

            self._align = value

    @property
    def allow_overlap(self) -> Optional[bool]:
        """If ``True``, annotation labels are allowed to overlap each other.

        Defaults to ``False``.

        :returns: Flag indicating whether to allow annotation labels to overlap.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_overlap

    @allow_overlap.setter
    def allow_overlap(self, value):
        if value is None:
            self._allow_overlap = None
        else:
            self._allow_overlap = bool(value)

    @property
    def background_color(self) -> Optional[str | Gradient | Pattern]:
        """The background color or gradient for the annotation's label. Defaults to
        ``'rgba(0, 0, 0, 0.75)'``.

        :returns: The backgorund color for the annotation's label.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        from highcharts_core import utility_functions
        self._background_color = utility_functions.validate_color(value)

    @property
    def border_color(self) -> Optional[str]:
        """The border color for the annotation's label. Defaults to
        ``'#000000'``.

        :returns: The border color for the annotation's label.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        self._border_color = validators.string(value, allow_empty = True)

    @property
    def border_radius(self) -> Optional[int | float | Decimal]:
        """The border radius (in pixels) applied to the annotation's label. Defaults to
        ``3``.

        :returns: The border radius to apply to the annotation's label.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = validators.numeric(value, allow_empty = True)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The border width (in pixels) applied to the annotation's label. Defaults to
        ``1``.

        :returns: The border width to apply to the annotation's label.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value, allow_empty = True)

    @property
    def class_name(self) -> Optional[str]:
        """A classname to apply styling using CSS. Defaults to
        ``'highcharts-no-tooltip'``.

        :returns: The classname to apply to enable styling via CSS.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def crop(self) -> Optional[bool]:
        """If ``True``, hide part of the annotation label that falls outside the plot
        area. Defaults to ``False``.

        :returns: Flag indicating whether to clip an annotation label that extends beyond
          the plot area.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._crop

    @crop.setter
    def crop(self, value):
        if value is None:
            self._crop = None
        else:
            self._crop = bool(value)

    @property
    def distance(self) -> Optional[int | float | Decimal]:
        """The label's distance in pixels from the point.

        :returns: The label's distance in pixels from the point.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._distance

    @distance.setter
    def distance(self, value):
        self._distance = validators.numeric(value,
                                            allow_empty = True)

    @property
    def format(self) -> Optional[str]:
        """A format string to apply to the label.

        .. seealso:

          * :meth:`PlotOptions.series.data_labels`

        :returns: The format string to apply to the labels.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._format

    @format.setter
    def format(self, value):
        self._format = validators.string(value, allow_empty = True)

    @property
    def formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format the annotation's label.

        Defaults to ``function () { return defined(this.y) ? this.y : 'Annotation label'; }``.

        .. note::

          If a :meth:`LabelOptions.format` or :meth:`LabelOptions.text` value are
          specified, the formatter will be ignored.

        .. hint::

          In the callback function, the use of ``this`` refers to a point object.

        :returns: A JavaScript callback function.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._formatter

    @formatter.setter
    @class_sensitive(CallbackFunction)
    def formatter(self, value):
        self._formatter = value

    @property
    def include_in_data_export(self) -> Optional[bool]:
        """If ``True``, the annotation is visible in the exported data table. Defaults to
        ``True``.

        :returns: Flag indicating whether to include the annotation label in the chart's
          exported data table.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._include_in_data_export

    @include_in_data_export.setter
    def include_in_data_export(self, value):
        if value is None:
            self._include_in_data_export = None
        else:
            self._include_in_data_export = bool(value)

    @property
    def overflow(self) -> Optional[str]:
        """Configuration on how to handle an annotation label that overflows outside of
        the plot area.  Defaults to ``'justify'``.

        Accepts:

          * ``'justify'`` - which forces the label back into the plot area
          * ``'none'`` - which applies no change

        .. note::

          The overflow treatment is also affected by the :meth:`LabelOptions.crop`
          setting.

        :returns: Configuration of overflow setting.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._overflow

    @overflow.setter
    def overflow(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._overflow = None
        else:
            value = value.lower()
            if value not in ['justify', 'none']:
                raise errors.HighchartsValueError(f'overflow accepts "justify" or "none".'
                                                  f' Was: {value}')
            self._overflow = value

    @property
    def padding(self) -> Optional[int]:
        """The padding within the border box when either
        :meth:`LabelOptions.border_width` or :meth:`LabelOptions.background_color` is set.

        Defaults to ``5``.

        :returns: The padding to apply to the annotation label.
        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._padding

    @padding.setter
    def padding(self, value):
        self._padding = validators.numeric(value, allow_empty = True)

    @property
    def shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration for the shadow to apply to the annotation box. Defaults to
        ``False``.

        If ``False``, no shadow is applied.

        :returns: The shadow configuration to apply or ``False``.
        :rtype: :class:`bool <python:bool>` or :class:`ShadowOptions` or
          :obj:`None <python:None>`
        """
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        if value is None:
            self._shadow = None
        elif isinstance(value, bool) and value is False:
            self._shadow = False
        else:
            value = validate_types(value,
                                   types = ShadowOptions)
            self._shadow = value

    @property
    def shape(self) -> Optional[str]:
        """The name of the symbol to use for the border around the label. Defaults to
        ``'callout'``.

        Accepts:

          * ``'connector'``
          * ``'rect'``
          * ``'circle'``
          * ``'diamond'``
          * ``'triangle'``
          * ``'callout'``

        :returns: The shape to use for the border around the label.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._shape

    @shape.setter
    def shape(self, value):
        if not value:
            self._shape = None
        else:
            value = validators.string(value, allow_empty = False)
            value = value.lower()
            if value not in ['callout',
                             'connector',
                             'rect',
                             'circle',
                             'diamond',
                             'triangle']:
                raise errors.HighchartsValueError(f'shape expects a supported annotation '
                                                  f'label shape. Was: {value}')
            self._shape = value

    @property
    def style(self) -> Optional[str]:
        """CSS styling to apply to the annotation's label.

        :rtype: :class:`str` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.string(value, allow_empty = True, coerce_value = True)

    @property
    def text(self) -> Optional[str]:
        """Alias for the :meth:`LabelOptions.format` property.

        :returns: The format string to apply to the labels.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._format

    @text.setter
    def text(self, value):
        self.format = value

    @property
    def use_html(self) -> Optional[bool]:
        """If ``True``, will use HTML to render the annotation's label. If ``False``, will
        use SVG or WebGL as applicable.

        Defaults to ``False``.

        :returns: Flag indicating whether to render annotation labels using HTML.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_html

    @use_html.setter
    def use_html(self, value):
        if value is None:
            self._use_html = None
        else:
            self._use_html = bool(value)

    @property
    def vertical_align(self) -> Optional[str]:
        """The vertical alignment of the annotation's label. Defaults to
        ``'bottom'``

        Accepts:

          * ``'bottom'``
          * ``'middle'``
          * ``'top'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._vertical_align

    @vertical_align.setter
    def vertical_align(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._vertical_align = None
        else:
            value = value.lower()
            if value not in ['bottom', 'middle', 'top']:
                raise errors.HighchartsValueError(f'vertical_align expects either "top", '
                                                  f'"middle", or "bottom". Was: {value}')
            self._vertical_align = value

    @property
    def x(self) -> Optional[int]:
        """The x position offset of the label relative to the point. Defaults to
        ``0``.

        .. note::

          If a :meth:`LabelOptions.distance` is defined, the distance takes precedence.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._x

    @x.setter
    def x(self, value):
        self._x = validators.numeric(value, allow_empty = True)

    @property
    def y(self) -> Optional[int]:
        """The y position offset of the label relative to the point. Defaults to
        ``-16``.

        .. note::

          If a :meth:`LabelOptions.distance` is defined, the distance takes precedence.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._y

    @y.setter
    def y(self, value):
        self._y = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'align': as_dict.get('align', None),
            'allow_overlap': as_dict.get('allowOverlap', None),
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'crop': as_dict.get('crop', None),
            'distance': as_dict.get('distance', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'overflow': as_dict.get('overflow', None),
            'padding': as_dict.get('padding', None),
            'shadow': as_dict.get('shadow', None),
            'shape': as_dict.get('shape', None),
            'style': as_dict.get('style', None),
            'text': as_dict.get('text', None),
            'use_html': as_dict.get('useHTML', None),
            'vertical_align': as_dict.get('verticalAlign', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'accessibility': self.accessibility,
            'align': self.align,
            'allowOverlap': self.allow_overlap,
            'backgroundColor': self.background_color,
            'borderColor': self.border_color,
            'borderRadius': self.border_radius,
            'borderWidth': self.border_width,
            'className': self.class_name,
            'crop': self.crop,
            'distance': self.distance,
            'format': self.format,
            'formatter': self.formatter,
            'includeInDataExport': self.include_in_data_export,
            'overflow': self.overflow,
            'padding': self.padding,
            'shadow': self.shadow,
            'shape': self.shape,
            'style': self.style,
            'text': self.text,
            'useHTML': self.use_html,
            'verticalAlign': self.vertical_align,
            'x': self.x,
            'y': self.y
        }

        return untrimmed


class AnnotationLabel(LabelOptions):
    """Configuration for an annotation label applied to a specific point.

    Used to override the global settings configured in :class:`LabelOptions` and applied
    via :meth:`Annotation.label_options`.

    """

    def __init__(self, **kwargs):
        self._point = None

        self.point = kwargs.get('point', None)

        super().__init__(**kwargs)

    @property
    def point(self) -> Optional[str | AnnotationPoint]:
        """Determines the point to which the label will be connected.

        It can be either the ID of the point which exists in the series, or a new point
        with defined x, y properties and optionally axes.

        :rtype: :class:`str <python:str>` or :class:`AnnotationPoint` or
          :obj:`None <python:None>`

        :raises HighchartsValueError: if cannot resolve the value to an allowed type
        """
        return self._point

    @point.setter
    def point(self, value):
        if not value:
            self._point = None
        elif isinstance(value, AnnotationPoint):
            self._point = value
        elif isinstance(value, str):
            try:
                self._point = AnnotationPoint.from_json(value)
            except ValueError:
                self._point = validators.string(value)
        elif isinstance(value, dict):
            self._point = AnnotationPoint.from_dict(value)
        else:
            raise errors.HighchartsValueError('Unable to resolve the value supplied to a '
                                              'supported type.')

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            # from LabelOptions
            'accessibility': as_dict.get('accessibility', None),
            'align': as_dict.get('align', None),
            'allow_overlap': as_dict.get('allowOverlap', None),
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'crop': as_dict.get('crop', None),
            'distance': as_dict.get('distance', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'overflow': as_dict.get('overflow', None),
            'padding': as_dict.get('padding', None),
            'shadow': as_dict.get('shadow', None),
            'shape': as_dict.get('shape', None),
            'style': as_dict.get('style', None),
            'text': as_dict.get('text', None),
            'use_html': as_dict.get('useHTML', None),
            'vertical_align': as_dict.get('verticalAlign', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),

            # from AnnotationLabel
            'point': as_dict.get('point', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'point': self.point
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


__all__ = [
    'LabelOptions',
    'AnnotationLabelOptionAccessibility',
    'ShadowOptions',
    'AnnotationLabel'
]



import tempfile
import os
from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class Data(HighchartsMeta):
    """The ``data`` property provides a simplified interface for adding data to a
    chart from sources like CVS, HTML tables, or grid views. See also
    `the tutorial article on the Data module <https://www.highcharts.com/docs/working-with-data/data-module>`_.

    .. warning::

      It requires the ``modules/data.js`` file to be loaded in the browser / client.

    .. warning::

      Please note that the default way of adding data in Highcharts, without the need
      of a module, is through the ``series.data`` property.

    """

    def __init__(self, **kwargs):
        self._before_parse = None
        self._columns = None
        self._columns_url = None
        self._complete = None
        self._csv = None
        self._csv_url = None
        self._data_refresh_rate = None
        self._date_format = None
        self._decimal_point = None
        self._enable_polling = None
        self._end_column = None
        self._end_row = None
        self._first_row_as_names = None
        self._google_api_key = None
        self._google_spreadsheet_key = None
        self._google_spreadsheet_range = None
        self._item_delimiter = None
        self._line_delimiter = None
        self._parsed = None
        self._parse_date = None
        self._rows = None
        self._rows_url = None
        self._series_mapping = None
        self._start_column = None
        self._start_row = None
        self._switch_rows_and_columns = None
        self._table = None

        self.before_parse = kwargs.get('before_parse', None)
        self.columns = kwargs.get('columns', None)
        self.columns_url = kwargs.get('columns_url', None)
        self.complete = kwargs.get('complete', None)
        self.csv = kwargs.get('csv', None)
        self.csv_url = kwargs.get('csv_url', None)
        self.data_refresh_rate = kwargs.get('data_refresh_rate', None)
        self.date_format = kwargs.get('date_format', None)
        self.decimal_point = kwargs.get('decimal_point', None)
        self.enable_polling = kwargs.get('enable_polling', None)
        self.end_column = kwargs.get('end_column', None)
        self.end_row = kwargs.get('end_row', None)
        self.first_row_as_names = kwargs.get('first_row_as_names', None)
        self.google_api_key = kwargs.get('google_api_key', None)
        self.google_spreadsheet_key = kwargs.get('google_spreadsheet_key', None)
        self.google_spreadsheet_range = kwargs.get('google_spreadsheet_range', None)
        self.item_delimiter = kwargs.get('item_delimiter', None)
        self.line_delimiter = kwargs.get('line_delimiter', None)
        self.parsed = kwargs.get('parsed', None)
        self.parse_date = kwargs.get('parse_date', None)
        self.rows = kwargs.get('rows', None)
        self.rows_url = kwargs.get('rows_url', None)
        self.series_mapping = kwargs.get('series_mapping', None)
        self.start_column = kwargs.get('start_column', None)
        self.start_row = kwargs.get('start_row', None)
        self.switch_rows_and_columns = kwargs.get('switch_rows_and_columns', None)
        self.table = kwargs.get('table', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'data'

    @property
    def before_parse(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function that is used to modify the CSV data before it is
        parsed. The function should return a modified version of the CSV string. Defaults
        to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._before_parse

    @before_parse.setter
    @class_sensitive(CallbackFunction)
    def before_parse(self, value):
        self._before_parse = value

    @property
    def columns(self) -> Optional[List[List[int | str | float | Decimal | type(None) | constants.EnforcedNullType]]]:
        """The data itself represented in a tabular form. Expects an iterable of
        iterables, where each second-level iterable represents a column of data. Defaults
        to :obj:`None <python:None>`.

        .. code-block:: python

          my_data = Data()
          my_data.columns = [
            [None, 'Apples', 'Pears', 'Oranges'], # categories
            ['Ola', 1, 4, 3], # first series for "Ola"
            ['Kari', 5, 4, 2] # second series for "Kari"
          ]

        .. note::

          Each cell in the second-level iterable can either be a
          :class:`str <python:str>`, numeric value, :obj:`None <python:None>`, or
          :class:`EnforcedNullType`.

        .. hint::

          Provided that :meth:`Data.switch_rows_and_columns` is not set, the columns will
          be interpreted as series.

        :rtype: :obj:`None <python:None>` or :class:`list <python:list>` of
          :class:`list <python:list>` instances with :class:`str <python:str>`, numeric,
          :obj:`None <python:None>`, or :class:`EnforcedNullType` members
        """
        return self._columns

    @columns.setter
    def columns(self, value):
        if not value:
            self._columns = None
        else:
            value = validators.iterable(value)
            processed_values = []
            for item in value:
                if not item or isinstance(item, constants.EnforcedNullType):
                    processed_values.append(item)
                elif isinstance(item, str):
                    processed_values.append(item)
                else:
                    item = validators.numeric(item)
                    processed_values.append(item)

            self._columns = processed_values

    @property
    def columns_url(self):
        """A URL to a remote JSON dataset, structured as a column array. Defaults to
        :obj:`None <python:None>`.

        .. note::

          Will be fetched when the chart is created using Ajax.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._columns_url

    @columns_url.setter
    def columns_url(self, value):
        if not value:
            self._columns_url = None
        else:
            try:
                self._columns_url = validators.url(value)
            except ValueError as error:
                try:
                    self._columns_url = validators.path(value)
                except ValueError:
                    raise error

    @property
    def complete(self) -> Optional[CallbackFunction]:
        """The JavaScript callback function that is evaluated when the data has finished
        loading (optionally from an external source) and been parsed. The first argument
        passed is a finished chart options object, containing the series. These options
        can be extended with additional options and passed directly to the chart
        constructor. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._complete

    @complete.setter
    @class_sensitive(CallbackFunction)
    def complete(self, value):
        self._complete = value

    @property
    def csv(self) -> Optional[str]:
        """A comma-delimited string to be parsed. Defaults to :obj:`None <python:None>`.

        .. seealso::

          The object has closely-related properties that determine which part of the CSV
          string should be used when constructing the chart and to configure how the CSV
          string should be parsed. In particular, please review:

            * :meth:`Data.start_row`
            * :meth:`Data.start_column`
            * :meth:`Data.end_row`
            * :meth:`Data.end_column`
            * :meth:`Data.line_delimiter`
            * :meth:`Data.item_delimiter`

        .. warning::

          The built-in CSV parser does *not* support all flavours of CSV, so in some cases
          it may be necessary to use an external CSV parser. See
          `this example <https://jsfiddle.net/highcharts/u59176h4/>`_ of parsing a CSV
          through the MIT-licensed `Papa Parse <http://papaparse.com/>`_ library.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._csv

    @csv.setter
    def csv(self, value):
        self._csv = validators.string(value, allow_empty = True)

    @property
    def csv_url(self):
        """A URL to a remote CSV dataset. Defaults to :obj:`None <python:None>`.

        .. note::

          Will be fetched when the chart is created using Ajax.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._csv_url

    @csv_url.setter
    def csv_url(self, value):
        if not value:
            self._csv_url = None
        else:
            try:
                self._csv_url = validators.url(value)
            except ValueError as error:
                try:
                    self._csv_url = validators.path(value)
                except ValueError:
                    raise error

    @property
    def data_refresh_rate(self) -> Optional[int | float | Decimal]:
        """The number of seconds between each poll of data from a remote source configured
        using either :meth:`Data.columns_url`, :meth:`Data.csv_url`,
        :meth:`Data.rows_url`, or :meth:`Data.google_spreadsheet_key`. Defaults to ``1``.

        .. warning::

          Cannot be set to less than ``1``.

        .. note::

          For polling to be enabled, :meth:`Data.enable_polling` must be ``True``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._data_refresh_rate

    @data_refresh_rate.setter
    def data_refresh_rate(self, value):
        self._data_refresh_rate = validators.numeric(value, allow_empty = True)

    @property
    def date_format(self) -> Optional[str]:
        """Indicates the format string to use to parse date values. If
        :obj:`None <python:None>`, defaults to a best-guess based on what format gives
        valid and ordered dates. Defaults to :obj:`None <python:None>`.

        Valid options include:

          * ``'YYYY/mm/dd'``
          * ``'dd/mm/YYYY'``
          * ``'mm/dd/YYYY'``
          * ``'dd/mm/YY'``
          * ``'mm/dd/YY'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._date_format

    @date_format.setter
    def date_format(self, value):
        self._date_format = validators.string(value, allow_empty = True)

    @property
    def decimal_point(self) -> Optional[str]:
        """The decimal point used when parsing number values. Defaults to ``'.'`` if
        :obj:`None <python:None>`.

        .. note::

          If both ``decimal_point`` and :meth:`Data.delimiter` are set to
          :obj:`None <python:None>`, the parser will attempt to deduce the decimal point
          automatically.

        :rtype: :class:`str <python:str>`
        """
        return self._decimal_point

    @decimal_point.setter
    def decimal_point(self, value):
        self._decimal_point = validators.string(value, allow_empty = True)

    @property
    def enable_polling(self) -> Optional[bool]:
        """If ``True``, automatically re-fetches remote datasets every *n* seconds (as
        per :meth:`Data.data_refresh_rate`). Defaults to ``False``.

        .. note::

          This flag only has an impact if remote datasets are in use, as specified by
          any one of:

            * :meth:`Data.columns_url`
            * :meth:`Data.csv_url`
            * :meth:`Data.google_spreadsheet_key`
            * :meth:`Data.rows_url`

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enable_polling

    @enable_polling.setter
    def enable_polling(self, value):
        if value is None:
            self._enable_polling = None
        else:
            self._enable_polling = bool(value)

    @property
    def end_column(self) -> Optional[int]:
        """In tabular input data, the last column (indexed by ``0``) to use. If
        :obj:`None <python:None>`, defaults to the last column that contains data.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._end_column

    @end_column.setter
    def end_column(self, value):
        self._end_column = validators.integer(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def end_row(self) -> Optional[int]:
        """In tabular input data, the last row (indexed by ``0``) to use. If
        :obj:`None <python:None>`, defaults to the last row that contains data.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._end_row

    @end_row.setter
    def end_row(self, value):
        self._end_row = validators.integer(value,
                                           allow_empty = True,
                                           minimum = 0)

    @property
    def first_row_as_names(self) -> Optional[bool]:
        """If ``True``, use the first row of data as series names. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._first_row_as_names

    @first_row_as_names.setter
    def first_row_as_names(self, value):
        if value is None:
            self._first_row_as_names = None
        else:
            self._first_row_as_names = bool(value)

    @property
    def google_api_key(self) -> Optional[str]:
        """The Google Spreadsheet API key required for access. Defaults to
        :obj:`None <python:None>`.

        .. note::

          Be sure to generate your Google Spreadsheet API key at Google's
          `API Services / Credentials <https://console.cloud.google.com/apis/credentials>`_.

        .. warning::

          To load data from Google Sheets, the :meth:`Data.google_spreadsheet_key` must be
          set, and must have access to the spreadsheet indicated by the spreadsheet key.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._google_api_key

    @google_api_key.setter
    def google_api_key(self, value):
        self._google_api_key = validators.string(value, allow_empty = True)

    @property
    def google_spreadsheet_key(self) -> Optional[str]:
        """The key or ``spreadsheetId`` value for the Google
        Sheets spreadsheet from which you wish to load data. Defaults to
        :obj:`None <python:None>`.

        .. hint::

          If you need help finding your spreadsheet key, please review the
          `Google Sheets API Overview <https://developers.google.com/sheets/api/guides/concepts>`_

        .. warning::

          To load data from Google Sheets, the :meth:`Data.google_api_key` must be set,
          and must have access to the spreadsheet indicated by the spreadsheet key.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._google_spreadsheet_key

    @google_spreadsheet_key.setter
    def google_spreadsheet_key(self, value):
        self._google_spreadsheet_key = validators.string(value, allow_empty = True)

    @property
    def google_spreadsheet_range(self) -> Optional[str]:
        """The Google Sheets A1 or R1C1 notation cell range from which to retrieve data.
        Defaults to :obj:`None <python:None>`.

        .. note::

          If set, this property takes precedence over :meth:`Data.start_column`,
          :meth:`Data.end_column`, :meth:`Data.start_row`, and :meth:`Data.end_row`.

        .. hint::

          For more details on how to determine and provide the spreadsheet range, please
          review the relevant
          `Google Sheets API documentation <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/get>`_.

        .. warning::

          To load data from Google Sheets, the :meth:`Data.google_api_key` and
          :meth:`Data.google_spreadsheet_key` must both be set.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._google_spreadsheet_range

    @google_spreadsheet_range.setter
    def google_spreadsheet_range(self, value):
        self._google_spreadsheet_range = validators.string(value, allow_empty = True)

    @property
    def item_delimiter(self) -> Optional[str | bool]:
        """Item or cell delimiter used when parsing CSV data. If
        :obj:`None <python:None>` or ``False``, defaults to the tab character ``\t`` if a
        tab character is found in the CSV string. If no tab character is found, defaults
        to ``','``.

        :rtype: :class:`str <python:str>` or :class:`bool <python:bool>` or
          :obj:`None <python:None>`
        """
        return self._item_delimiter

    @item_delimiter.setter
    def item_delimiter(self, value):
        if value is False:
            self._item_delimiter = False
        else:
            self._item_delimiter = validators.string(value, allow_empty = True)

    @property
    def line_delimiter(self) -> Optional[str]:
        """The string used to delimit records (lines) when parsing CSV data. Defaults to
        :obj:`None <python:None>`, which assumes ``'\\n'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._line_delimiter

    @line_delimiter.setter
    def line_delimiter(self, value):
        self._line_delimiter = validators.string(value, allow_empty = True)

    @property
    def parsed(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function to access the parsed columns (the
        two-dimentional input data array) directly *before* they are interpreted into
        series data and categories. The function should return ``false`` to stop
        completion, or call (in JavaScript) ``this.complete()`` to continue
        asynchronously. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._parsed

    @parsed.setter
    @class_sensitive(CallbackFunction)
    def parsed(self, value):
        self._parsed = value

    @property
    def parse_date(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function to parse string representations of dates into
        JavaScript timestamps. Should return an integer timestamp on success. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._parse_date

    @parse_date.setter
    @class_sensitive(CallbackFunction)
    def parse_date(self, value):
        self._parse_date = value

    @property
    def rows(self) -> Optional[List[List[int | str | float | Decimal | type(None) | constants.EnforcedNullType]]]:
        """The data itself represented in a tabular form. Expects an iterable of
        iterables, where each second-level iterable represents a row of data. Defaults
        to :obj:`None <python:None>`.

        .. code-block:: python

          my_data = Data()
          my_data.rows = [
            [None, 'Apples', 'Pears', 'Oranges'], # categories
            ['Ola', 1, 4, 3], # first series for "Ola"
            ['Kari', 5, 4, 2] # second series for "Kari"
          ]

        .. note::

          Each cell in the second-level iterable can either be a
          :class:`str <python:str>`, numeric value, :obj:`None <python:None>`, or
          :class:`EnforcedNullType`.

        .. hint::

          Provided that :meth:`Data.switch_rows_and_columns` is not set, the rows will
          be interpreted as series.

        :rtype: :obj:`None <python:None>` or :class:`list <python:list>` of
          :class:`list <python:list>` instances with :class:`str <python:str>`, numeric,
          :obj:`None <python:None>`, or :class:`EnforcedNullType` members
        """
        return self._rows

    @rows.setter
    def rows(self, value):
        if not value:
            self._rows = None
        else:
            value = validators.iterable(value)
            processed_values = []
            for item in value:
                if not item or isinstance(item, constants.EnforcedNullType):
                    processed_values.append(item)
                elif isinstance(item, str):
                    processed_values.append(item)
                else:
                    item = validators.numeric(item)
                    processed_values.append(item)

            self._rows = processed_values

    @property
    def rows_url(self):
        """A URL to a remote JSON dataset, structured as a row array. Defaults to
        :obj:`None <python:None>`.

        .. note::

          Will be fetched when the chart is created using Ajax.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._rows_url

    @rows_url.setter
    def rows_url(self, value):
        if not value:
            self._rows_url = None
        else:
            try:
                self._rows_url = validators.url(value)
            except ValueError as error:
                try:
                    self._rows_url = validators.path(value)
                except ValueError:
                    raise error

    @property
    def series_mapping(self) -> Optional[dict[str, int]]:
        """A dictionary used to indicate where point properties are to be found in the data.
        Property names are the keys, while the values are the column indices indicating where
        in the CSV the data for that point is to be found. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`dict <python:dict>` with :class:`str <python:str>` keys and
          :class:`int <python:int>` values, or :obj:`None <python:None>`.
        """
        return self._series_mapping

    @series_mapping.setter
    def series_mapping(self, value):
        if not value:
            self._series_mapping = None
        else:
            value = validators.dict(value)
            for key in value:
                key = validators.string(value)
                item = value[key]
                item = validators.integer(item)

            self._series_mapping = value

    @property
    def start_column(self) -> Optional[int]:
        """In tabular input data, the first column (indexed by ``0``) to use. If
        :obj:`None <python:None>`, defaults to ``0``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._start_column

    @start_column.setter
    def start_column(self, value):
        self._start_column = validators.integer(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def start_row(self) -> Optional[int]:
        """In tabular input data, the first row (indexed by ``0``) to use. If
        :obj:`None <python:None>`, defaults to ``0``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._start_row

    @start_row.setter
    def start_row(self, value):
        self._start_row = validators.integer(value,
                                             allow_empty = True,
                                             minimum = 0)

    @property
    def switch_rows_and_columns(self) -> Optional[bool]:
        """If ``True``, swiches the interpretation of columns and rows so that
        :meth:`Data.columns` effectively becomes rows of the data while :meth:`Data.rows`
        is interpreted as the series of data. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._switch_rows_and_columns

    @switch_rows_and_columns.setter
    def switch_rows_and_columns(self, value):
        if value is None:
            self._switch_rows_and_columns = None
        else:
            self._switch_rows_and_columns = bool(value)

    @property
    def table(self) -> Optional[str]:
        """An HTML table (in :class:`str <pythoN:str>` form), or the ID of such a table
        found in the browser DOM, which should be parsed to extract the data intended for
        visualization. Defaults to :obj:`None <python:None>`.

        .. seealso::

          The object has closely-related properties that determine which part of the HTML
          table should be used when constructing the chart and to configure how the table
          string should be parsed. In particular, please review:

            * :meth:`Data.start_row`
            * :meth:`Data.start_column`
            * :meth:`Data.end_row`
            * :meth:`Data.end_column`

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._table

    @table.setter
    def table(self, value):
        self._table = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'before_parse': as_dict.get('beforeParse', None),
            'columns': as_dict.get('columns', None),
            'columns_url': as_dict.get('columnsURL', None),
            'complete': as_dict.get('complete', None),
            'csv': as_dict.get('csv', None),
            'csv_url': as_dict.get('csvURL', None),
            'data_refresh_rate': as_dict.get('dataRefreshRate', None),
            'date_format': as_dict.get('dateFormat', None),
            'decimal_point': as_dict.get('decimalPoint', None),
            'enable_polling': as_dict.get('enablePolling', None),
            'end_column': as_dict.get('endColumn', None),
            'end_row': as_dict.get('endRow', None),
            'first_row_as_names': as_dict.get('firstRowAsNames', None),
            'google_api_key': as_dict.get('googleAPIKey', None),
            'google_spreadsheet_key': as_dict.get('googleSpreadsheetKey', None),
            'google_spreadsheet_range': as_dict.get('googleSpreadsheetRange', None),
            'item_delimiter': as_dict.get('itemDelimiter', None),
            'line_delimiter': as_dict.get('lineDelimiter', None),
            'parsed': as_dict.get('parsed', None),
            'parse_date': as_dict.get('parseDate', None),
            'rows': as_dict.get('rows', None),
            'rows_url': as_dict.get('rowsURL', None),
            'series_mapping': as_dict.get('seriesMapping', None),
            'start_column': as_dict.get('startColumn', None),
            'start_row': as_dict.get('startRow', None),
            'switch_rows_and_columns': as_dict.get('switchRowsAndColumns', None),
            'table': as_dict.get('table', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'beforeParse': self.before_parse,
            'columns': self.columns,
            'columnsURL': self.columns_url,
            'complete': self.complete,
            'csv': self.csv,
            'csvURL': self.csv_url,
            'dataRefreshRate': self.data_refresh_rate,
            'dateFormat': self.date_format,
            'decimalPoint': self.decimal_point,
            'enablePolling': self.enable_polling,
            'endColumn': self.end_column,
            'endRow': self.end_row,
            'firstRowAsNames': self.first_row_as_names,
            'googleAPIKey': self.google_api_key,
            'googleSpreadsheetKey': self.google_spreadsheet_key,
            'googleSpreadsheetRange': self.google_spreadsheet_range,
            'itemDelimiter': self.item_delimiter,
            'lineDelimiter': self.line_delimiter,
            'parsed': self.parsed,
            'parseDate': self.parse_date,
            'rows': self.rows,
            'rowsURL': self.rows_url,
            'seriesMapping': self.series_mapping,
            'startColumn': self.start_column,
            'startRow': self.start_row,
            'switchRowsAndColumns': self.switch_rows_and_columns,
            'table': self.table,
        }

        return untrimmed

    @classmethod
    def from_pandas(cls,
                    as_df,
                    represent_as = 'csv',
                    data_kwargs = None,
                    pandas_kwargs = None):
        """Create a :class:`Data` instance from a Pandas
        :class:`DataFrame <pandas:DataFrame`.

        :param as_df: The :class:`DataFrame <pandas:DataFrame>` from which to create the
          :class:`Data` instance.
        :type as_df: :class:`DataFrame <pandas:DataFrame>`

        :param represent_as: The format to which ``as_df`` should be serialized. Accepts
          ``'csv'`` or ``'html'``. Defaults to ``'csv'``.
        :type represent_as: :class:`str <python:str>`

        :param data_kwargs: Additional keyword arguments to pass to the :class:`Data`
          constructor (``__init__()``) method. Defaults to :obj:`None <python:None>`.
        :type data_kwargs: :class:`dict <python:dict>`

        :param pandas_kwargs: Keyword arguments to pass to the Pandas
          :meth:`DataFrame.to_csv()` or :meth:`DataFrame.to_html()` methods. Defaults to
          :obj:`None <python:None>`.
        :type pandas_kwargs: :class:`dict <python:dict>`

        .. note::

          To prevent unexpected behavior, if ``represent_as`` is set to ``'csv'``, the
          :meth:`Data.table` property will be set to :obj:`None <python:None>`. If
          ``represent_as`` is set to ``'html'``, the :meth:`Data.csv` property will be set
          to :obj:`None <python:None>`.

        :returns: A :class:`Data` instance.
        """
        from pandas import DataFrame

        if not data_kwargs:
            data_kwargs = {}
        if not pandas_kwargs:
            pandas_kwargs = {}

        if not isinstance(as_df, DataFrame):
            raise errors.HighchartsValueError(f'as_df must be a Pandas DataFrame.'
                                              f'Was: {as_df.__class__.__name__}')

        represent_as = validators.string(represent_as, allow_empty = True) or 'csv'
        represent_as = represent_as.lower()
        if represent_as not in ['csv', 'html']:
            raise errors.HighchartsValueError(f'represent_as expects either "csv" or '
                                              f'"html". Received: {represent_as}')
        if represent_as == 'csv':
            pandas_kwargs['path_or_buf'] = None
            as_csv = as_df.to_csv(**pandas_kwargs)

            data_kwargs['csv'] = as_csv
            data_kwargs['table'] = None
        else:
            pandas_kwargs['buf'] = None
            as_table = as_df.to_html(**pandas_kwargs)

            data_kwargs['csv'] = None
            data_kwargs['table'] = as_table

        return cls(**data_kwargs)

    @classmethod
    def from_pyspark(cls,
                     as_df,
                     data_kwargs = None,
                     pyspark_kwargs = None,
                     consolidation = 'repartition'):
        """Create a :class:`Data` instance from a PySpark
        :class:`DataFrame <pyspark:sql.DataFrame>`.

        .. warning::

          If you are using PySpark, you might be working with extremely large datasets
          (proverbial "big data"). Are you sure you want to be visualizing datasets of
          such size? If so, you should be aware that Highcharts for Python tries to store
          the dataset in the Python environment's memory - **not** on disk.

          If you need to read large datasets from disk, using the
          :meth:`Data.from_pyspark()` is **not** recommended for this reason.


        :param as_df: The :class:`DataFrame <pyspark:sql.DataFrame>` from which to create
          the :class:`Data` instance.
        :type as_df: :class:`pyspark.sql.DataFrame <pyspark:sql.DataFrame>`

        :param data_kwargs: Additional keyword arguments to pass to the :class:`Data`
          constructor (``__init__()``) method. Defaults to :obj:`None <python:None>`.

          .. warning::

            The :meth:`Data.csv` and :meth:`Data.table` properties will be overwritten
            by Highcharts for Python if specified, so don't bother.

        :type data_kwargs: :class:`dict <python:dict>`

        :param pyspark_kwargs: Keyword arguments to pass to the Pyspark
          :meth:`DataFrame.write().csv() <pyspark:sql.DataFrameWriter.csv>` method.
          Defaults to :obj:`None <python:None>`.

          .. warning::

            The ``path`` and ``mode`` arguments will be overwritten by Highcharts for
            Python if specified, so don't bother.

        :type pyspark_kwargs: :class:`dict <python:dict>`

        :param consolidation: If ``'repartition'``, will repartition your PySpark
          :class:`DataFrame <pyspark:pyspark.sql.DataFrame>` into a single unpartitioned
          table prior to the generation of a CSV dataset. If ``'coalesce'``, will
          coalesce your PySpark :class:`DataFrame <pyspark:pyspark.sql.DataFrame>` into
          a single unpartitioned table prior to the generation of a CSV dataset. If
          :obj:`None <python:None>`, will not apply any consolidation to the
          :class:`DataFrame <pyspark:pyspark.sql.DataFrame>`. Defaults to ``'coalesce'``.

          .. hint::

            Setting this value to ``'coalesce'`` is particularly useful if you are working
            on your :class:`DataFrame <pyspark:pyspark.sql.DataFrame>` in multiple Spark
            nodes to prevent loss of data.

            Setting this value to :obj:`None <python:None>` may provide a boost to
            performance, however use with caution as it may lead to unexpected data loss
            or errors if using multiple Spark nodes.

        :returns: A :class:`Data` instance.
        """
        from pyspark.sql import DataFrame

        if not data_kwargs:
            data_kwargs = {}
        if not pyspark_kwargs:
            pyspark_kwargs = {}
            pyspark_kwargs['mode'] = 'overwrite'

        if not isinstance(as_df, DataFrame):
            raise errors.HighchartsValueError(f'as_df must be a Pyspark DataFrame.'
                                              f'Was: {as_df.__class__.__name__}')

        with tempfile.NamedTemporaryFile() as csv_file:
            pyspark_kwargs['path'] = csv_file.name

            if consolidation == 'repartition':
                as_df.repartition(1).write.csv(**pyspark_kwargs)
            elif consolidation:
                as_df.coalesce(1).write.csv(**pyspark_kwargs)
            else:
                as_df.write.csv(**pyspark_kwargs)

            csv_file.seek(0)
            data_kwargs['csv'] = csv_file.read()

        data_kwargs['table'] = None

        return cls(**data_kwargs)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.metaclasses import HighchartsMeta


class Caption(HighchartsMeta):
    """The chart's caption, which will render below the chart and will be part of
    exported charts.

    .. note::

      The caption can be updated after chart initialization in JavaScript through the
      ``Chart.update()`` or ``Chart.caption.update()`` methods.

    """

    def __init__(self, **kwargs):
        self._align = None
        self._floating = None
        self._margin = None
        self._style = None
        self._text = None
        self._use_html = None
        self._vertical_align = None
        self._x = None
        self._y = None

        self.align = kwargs.get('align', None)
        self.floating = kwargs.get('floating', None)
        self.margin = kwargs.get('margin', None)
        self.style = kwargs.get('style', None)
        self.text = kwargs.get('text', None)
        self.use_html = kwargs.get('use_html', None)
        self.vertical_align = kwargs.get('vertical_align', None)
        self.x = kwargs.get('x', None)
        self.y = kwargs.get('y', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'caption'

    @property
    def align(self) -> Optional[str]:
        """The alignment of the caption. Defaults to
        ``'left'``.

        Accepts:

          * ``'left'``
          * ``'center'``
          * ``'right'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._align

    @align.setter
    def align(self, value):
        if not value:
            self._align = None
        else:
            value = validators.string(value, allow_empty = False)
            value = value.lower()
            if value not in ['left', 'center', 'right']:
                raise errors.HighchartsValueError(f'align must be either "left", '
                                                  f'"center", or "right". Was: {value}')

        self._align = value

    @property
    def floating(self) -> Optional[bool]:
        """If ``True`, sets the caption to floating. When the caption is floating, the
        plot area will not move to make space for it. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._floating

    @floating.setter
    def floating(self, value):
        if value is None:
            self._floating = None
        else:
            self._floating = bool(value)

    @property
    def margin(self) -> Optional[int | float | Decimal]:
        """The margin between the caption and the plot area. Defaults to
        ``15``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._margin

    @margin.setter
    def margin(self, value):
        self._margin = validators.numeric(value, allow_empty = True)

    @property
    def style(self) -> Optional[str | dict]:
        """CSS styling to apply to the caption. Defaults to
        ``{"color": "#666666", "fontSize": "0.8em"}``.

        :rtype: :class:`dict <python:dict>` or :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        try:
            self._style = validators.dict(value, allow_empty = True)
        except (ValueError, TypeError):
            self._style = validators.string(value, 
                                            allow_empty = True,
                                            coerce_value = True)

    @property
    def text(self) -> Optional[str]:
        """The caption text of the chart. Defaults to an empty string (``''``).

        :rtype: :class:`str <python:str>`
        """
        return self._text or ''

    @text.setter
    def text(self, value):
        self._text = validators.string(value, allow_empty = True)

    @property
    def use_html(self) -> Optional[bool]:
        """If ``True``, will use HTML to render the caption. If ``False``, will
        use SVG or WebGL as applicable.

        Defaults to ``False``.

        :returns: Flag indicating whether to render the caption using HTML.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_html

    @use_html.setter
    def use_html(self, value):
        if value is None:
            self._use_html = None
        else:
            self._use_html = bool(value)

    @property
    def vertical_align(self) -> Optional[str]:
        """The vertical alignment of the caption. Defaults to
        ``'bottom'``.

        Accepts:

          * ``'bottom'``
          * ``'middle'``
          * ``'top'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._vertical_align

    @vertical_align.setter
    def vertical_align(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._vertical_align = None
        else:
            value = value.lower()
            if value not in ['bottom', 'middle', 'top']:
                raise errors.HighchartsValueError(f'vertical_align expects either "top", '
                                                  f'"middle", or "bottom". Was: {value}')
            self._vertical_align = value

    @property
    def x(self) -> Optional[int | float | Decimal]:
        """The x position of the caption relative to the alignment within
        :meth:`Options.chart.spacing_left` and :meth:`Option.chart.spacing_right`.
        Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._x

    @x.setter
    def x(self, value):
        self._x = validators.numeric(value, allow_empty = True)

    @property
    def y(self) -> Optional[int]:
        """The y position of the caption relative to the alignment within
        :meth:`Options.chart.spacing_left` and :meth:`Option.chart.spacing_right`.
        Defaults to ``None``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._y

    @y.setter
    def y(self, value):
        self._y = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'align': as_dict.get('align', None),
            'floating': as_dict.get('floating', None),
            'margin': as_dict.get('margin', None),
            'style': as_dict.get('style', None),
            'text': as_dict.get('text', None),
            'use_html': as_dict.get('useHTML', None),
            'vertical_align': as_dict.get('verticalAlign', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'align': self.align,
            'floating': self.floating,
            'margin': self.margin,
            'style': self.style,
            'text': self.text,
            'useHTML': self.use_html,
            'verticalAlign': self.vertical_align,
            'x': self.x,
            'y': self.y
        }

        return untrimmed
